"""Main WaterSmart module"""

import aiohttp
import asyncio
import async_timeout
import logging
import socket
import time

from aiohttp_client_cache import CachedSession, SQLiteBackend
from importlib.metadata import version


class WatersmartClientError(Exception):
    """Exception to indicate a general API error."""


class WatersmartClientCommunicationError(WatersmartClientError):
    """Exception to indicate a communication error."""


class WatersmartClientAuthenticationError(WatersmartClientError):
    """Exception to indicate an authentication error."""


class WatersmartClient:
    def __init__(self, *, url, email, password, session=None):
        self._url = url
        self._email = email
        self._password = password
        self._data_series = []
        if session:
            self._session = session
        else:
            self._headers = {"User-Agent": "py-watersmart " + version("py-watersmart")}
            self._cache = SQLiteBackend(
                expire_after=60 * 60 * 6,
                include_headers=False,
                cache_name="~/.cache/py-watersmart.db",
            )
            self._session = CachedSession(
                cache=self._cache,
                headers=self._headers,
            )
        assert "watersmart.com" in url, "Expected a watersmart.com URL"
        assert "http" in url, "Expected an http/https schema"
        logging.debug("WatersmartClient ready, headers: %s", self._headers)

    async def _login(self):
        url = f"{self._url}/index.php/welcome/login?forceEmail=1"
        login = {"token": "", "email": self._email, "password": self._password}
        result = await self._session.post(url, data=login)
        logging.debug(result)

    async def _populate_data(self):
        url = f"{self._url}/index.php/rest/v1/Chart/RealTimeChart"
        chart_rsp = await self._session.get(url)
        if chart_rsp.status != 200:
            raise WatersmartClientAuthenticationError()
        data = await chart_rsp.json()
        self._data_series = data["data"]["series"]

    @classmethod
    def _amend_with_local_ts(cls, datapoint):
        # The read_datetime is a timestamp in local TZ, not UTC, which
        # confuses python datetime, so use the naive struct_time
        ts = time.gmtime(datapoint["read_datetime"])
        result = datapoint.copy()
        result["local_datetime"] = time.strftime("%Y-%m-%d %H:%M:%S", ts)
        return result

    async def usage(self):
        if not self._data_series:
            try:
                async with async_timeout.timeout(10):
                    logging.debug("Loading watersmart data from %s", self._url)
                    await self._login()
                    await self._populate_data()
            except WatersmartClientAuthenticationError as e:
                raise e
            except asyncio.TimeoutError as exception:
                raise WatersmartClientCommunicationError(
                    "Timeout error fetching information",
                ) from exception
            except (aiohttp.ClientError, socket.gaierror) as exception:
                raise WatersmartClientCommunicationError(
                    "Error fetching information",
                ) from exception
            except Exception as exception:  # pylint: disable=broad-except
                raise WatersmartClientError(
                    "Something really wrong happened!"
                ) from exception
            finally:
                await self._close()

        result = []
        for datapoint in self._data_series:
            result.append(WatersmartClient._amend_with_local_ts(datapoint))
        return result

    async def _close(self):
        await self._session.close()



"""Download Watersmart meter data"""

import asyncio
import argparse
import logging
from watersmart import (
    WatersmartClient,
    WatersmartClientAuthenticationError,
    WatersmartClientCommunicationError,
    WatersmartClientError,
)


PARSER = argparse.ArgumentParser(description=__doc__)

PARSER.add_argument(
    "--log-level",
    default=logging.INFO,
    type=lambda x: getattr(logging, x.upper()),
    help="Configure the logging level.",
)
PARSER.add_argument("--url", required=True)
PARSER.add_argument("--email", required=True)
PARSER.add_argument("--password", required=True)


async def main():
    args = PARSER.parse_args()
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=args.log_level, format=log_format)
    wc = WatersmartClient(url=args.url, email=args.email, password=args.password)
    try:
        data = await wc.usage()
        for datapoint in sorted(data, key=lambda x: x["read_datetime"]):
            parts = [
                f"{datapoint['local_datetime']}",
                f"usage: {datapoint['gallons']:8}gal",
                f"leak: {datapoint['leak_gallons']:8}gal",
                f"flags: {datapoint['flags']}",
            ]
            print(" | ".join(parts))
    except WatersmartClientAuthenticationError:
        logging.exception("Login failure")
    except WatersmartClientCommunicationError:
        logging.exception("Communications error")
    except WatersmartClientError:
        logging.exception("Unknown error")


def start():
    asyncio.run(main())


if __name__ == "__main__":
    start()



#!/usr/bin/python3

import pytest
import os
import time

from . import WatersmartClient

pytest_plugins = ("pytest_asyncio",)


def _assert_amend_with_local_ts():
    datapoint = {
        "read_datetime": 1717488000,
        "gallons": 67.2,
        "flags": None,
        "leak_gallons": 0,
    }
    result = WatersmartClient._amend_with_local_ts(datapoint)
    assert result["read_datetime"] == datapoint["read_datetime"]
    assert result["gallons"] == datapoint["gallons"]
    assert result["leak_gallons"] == datapoint["leak_gallons"]
    assert result["flags"] == datapoint["flags"]
    assert "local_datetime" not in datapoint
    assert "local_datetime" in result
    assert f"{result['local_datetime']}" == "2024-06-04 08:00:00"


def test_amend_with_local_ts_in_local():
    _assert_amend_with_local_ts()


def test_amend_with_local_ts_in_pacific():
    os.environ["TZ"] = "America/Los_Angeles"
    time.tzset()
    _assert_amend_with_local_ts()


def test_amend_with_local_ts_in_utc():
    os.environ["TZ"] = "utc"
    time.tzset()
    _assert_amend_with_local_ts()


@pytest.mark.asyncio
async def test_disallow_non_watersmart_urls():
    with pytest.raises(AssertionError):
        WatersmartClient(
            url="https://wsmart.example", email="e@mail", password="passw4rd"
        )
